#include "exi.h"
#include <irq.h>
#include <console.h>
#include "cache.h"

static int flags;

/* exi_select: enable chip select, set speed */
void exi_select(int channel, int device, int freq)
{
	volatile unsigned long *exi = (volatile unsigned long *)0xCC006800;
	long d;
	cli(flags);
	// exi_select
	d = exi[channel * 5];
	d &= 0x405;
	d |= ((1<<device)<<7) | (freq << 4);
	exi[channel*5] = d;
}

/* disable chipselect */
void exi_deselect(int channel)
{
	volatile unsigned long *exi = (volatile unsigned long *)0xCC006800;
	exi[channel * 5] &= 0x405;
	sti(flags);
}

/* dirty way for asynchronous reads */
static void *exi_last_addr;
static int   exi_last_len;

/* mode?Read:Write len bytes to/from channel */
/* when read, data will be written back in exi_sync */
inline void exi_imm(int channel, void *data, int len, int mode, int zero)
{
	volatile unsigned long *exi = (volatile unsigned long *)0xCC006800;
	if (mode == EXI_WRITE)
		exi[channel * 5 + 4] = *(unsigned long*)data;
	else
		exi[channel * 5 + 4] = -1;

	exi[channel * 5 + 3] = ((len-1)<<4)|(mode<<2)|1;
	if (mode == EXI_READ)
	{
		exi_last_addr = data;
		exi_last_len = len;
	} else
	{
		exi_last_addr = 0;
		exi_last_len = 0;
	}
}

/* Wait until transfer is done, write back data */
void exi_sync(int channel)
{
	volatile unsigned long *exi = (volatile unsigned long *)0xCC006800;
	while (exi[channel * 5 + 3] & 1);

	if (exi_last_addr)
	{	
		int i;
		unsigned long d;
		d = exi[channel * 5 + 4];
		for (i=0; i<exi_last_len; ++i)
			((unsigned char*)exi_last_addr)[i] = (d >> ((3-i)*8)) & 0xFF;
	}
}

/* simple wrapper for transfers > 4bytes */
void exi_imm_ex(int channel, void *data, int len, int mode)
{
	unsigned char *d = (unsigned char*)data;
	while (len)
	{
		int tc = len;
		if (tc > 4)
			tc = 4;
		exi_imm(channel, d, tc, mode, 0);
		exi_sync(channel);
		len-=tc;
		d+=tc;
	}
}

void exi_dma_start(int channel, void *data, int len, int mode)
{
	volatile unsigned long *exi = (volatile unsigned long *)0xCC006800;
	if (len < 0)
	{
		printf("!!! len is %d\n", len);
		while (1);
	}
		
	exi[channel * 5 + 1] = (unsigned long)data;
	exi[channel * 5 + 2] = len;
	exi[channel * 5 + 3] = (mode<<2)|3;
}

void exi_imm_fast(int channel, void *data, int len, int mode)
{
	unsigned long addr = (unsigned long)data;
	unsigned char *d = (unsigned char*)data;
	
	int addr_misalign = (-addr) & 31;

	if ((len - addr_misalign) < 32)
		return exi_imm_ex(channel, data, len, mode);

	if (addr_misalign) /* fix misalignment with normal transfers */
	{
		exi_imm_ex(channel, data, addr_misalign, mode);
		len  -= addr_misalign;
		addr += addr_misalign;
	}
	
	int len_align = len & ~31;
	
	dcache_flush((void*)addr, len_align);
	exi_dma_start(channel, (void*)addr, len_align, mode);
	exi_sync(channel);
	if (mode == EXI_READ)
		dcache_inv((void*)addr, len_align);
	
	len -= len_align;
	if (len)
		exi_imm_ex(channel, (void*)(addr + len_align), len, mode);
}

static exi_irq_handler_t *exi_handler[3 * 3]; // 3 channels, 3 events
static void *exi_handler_context[3 * 3];
static unsigned long exi_enable_mask[3];

void exi_refresh_enable()
{
	int channel;
	for (channel = 0; channel < 3; ++channel)
		*(unsigned long*)(0xcc006800 + channel*0x14) |= 0x401; // exi_enable_mask[channel];
}

static inline int have_irq(int t)
{
	if (exi_handler[t])
	{
		exi_handler[t](t/3, t%3, exi_handler_context[t]);
		return 1;
	} else
	{
		printf("unhandled exi irq!\n");
		return 0;
	}
}

int exi_interrupt_handler(int irq, void *c)
{
	int ch;
	for (ch = 0; ch < 3; ++ch)
	{
		unsigned long v = (*(unsigned long*)(0xcc006800 + ch * 0x14)); //  & 0xC0F;
		v &= v<<1;
		*(unsigned long*)(0xcc006800 + ch * 0x14) |= v;
		if (v & 0x800)
			have_irq(ch * 3 + EXI_EVENT_INSERT);
		if (v & 8)
			have_irq(ch * 3 + EXI_EVENT_TC);
		if (v & 2)
			have_irq(ch * 3 + EXI_EVENT_IRQ);
	}
	return 1;
}

void exi_init()
{
	*(unsigned long*)(0xcc006800) |= 1<<13; // disable decryption and rom mapping
	
	printf("EXI: init interrupts\n");
	request_irq(IRQ_EXI, exi_interrupt_handler, 0);
	
	*(unsigned long*)(0xcc006814) |= 3<<10; // enable&clear irq for insertion
}

int exi_request_irq(int channel, int event, exi_irq_handler_t *handler, void *context)
{
	if (exi_handler[channel * 3 + event])
	{
		printf("EXI: irq %d:%d already used!\n", channel, event);
		return -1;
	}
	exi_handler[channel * 3 + event] = handler;
	exi_handler_context[channel * 3 + event] = context;
	
	switch (event)
	{
	case EXI_EVENT_TC:
		*(unsigned long*)(0xcc006800 + channel*0x14) |= 3<<2;  // enable&clear irq
		exi_enable_mask[channel] |= 1 << 2;
		break;
	case EXI_EVENT_INSERT:
		*(unsigned long*)(0xcc006800 + channel*0x14) |= 3<<10;
		exi_enable_mask[channel] |= 1 << 10;
		break;
	case EXI_EVENT_IRQ:
		*(unsigned long*)(0xcc006800 + channel*0x14) |= 3<<0;
		exi_enable_mask[channel] |= 1 << 0;
		break;
	}
	return 0;
}

int exi_free_irq(int channel, int event)
{
	exi_handler[channel * 3 + event] = 0;
	switch (event)
	{
	case EXI_EVENT_TC:
		*(unsigned long*)(0xcc006800 + channel*0x14) &= ~(3<<2);  // enable&clear irq
		exi_enable_mask[channel] &= ~(1 << 2);
		break;
	case EXI_EVENT_INSERT:
		*(unsigned long*)(0xcc006800 + channel*0x14) &= ~(3<<10);
		exi_enable_mask[channel] &= ~(1 << 10);
		break;
	case EXI_EVENT_IRQ:
		*(unsigned long*)(0xcc006800 + channel*0x14) &= ~(3<<0);
		exi_enable_mask[channel] &= ~(1 << 0);
		break;
	}
	return 0;
}

void exi_read(int channel, void *data, int len)
{
	exi_imm_ex(channel, data, len, EXI_READ);
}

void exi_write(int channel, const void *data, int len)
{
	exi_imm_ex(channel, data, len, EXI_WRITE);
}

unsigned long exi_rw(int channel,unsigned long data,int len)
{
	volatile unsigned long *exi = (volatile unsigned long *)0xCC006800;
	exi[channel * 5 + 4] = data<<(8*(4-len));
	exi[channel * 5 + 3] = ((len-1)<<4)|(2<<2)|1;
	while (exi[channel * 5 + 3] & 1);
	return exi[channel * 5 + 4]>>(8*(4-len));
}
