#include "lwip/opt.h"
#include "lwip/def.h"
#include "lwip/err.h"
#include "lwip/mem.h"
#include "lwip/pbuf.h"
#include "lwip/stats.h"
#include "lwip/sys.h"
#include "netif/etharp.h"
#include "lwip/snmp.h"

#include "time.h"
#include "irq.h"

#include "exi.h"

#define ETH_ZLEN  60

/*
 * Macronix mx98728ec supporting bits.
 *
 */

#define BBA_NCRA 0x00		/* Network Control Register A, RW */
#define   BBA_NCRA_RESET     (1<<0)	/* RESET */
#define   BBA_NCRA_ST0       (1<<1)	/* ST0, Start transmit command/status */
#define   BBA_NCRA_ST1       (1<<2)	/* ST1,  " */
#define   BBA_NCRA_SR        (1<<3)	/* SR, Start Receive */

#define BBA_NCRB 0x01		/* Network Control Register B, RW */
#define   BBA_NCRB_PR        (1<<0)	/* PR, Promiscuous Mode */
#define   BBA_NCRB_CA        (1<<1)	/* CA, Capture Effect Mode */
#define   BBA_NCRB_PM        (1<<2)	/* PM, Pass Multicast */
#define   BBA_NCRB_PB        (1<<3)	/* PB, Pass Bad Frame */
#define   BBA_NCRB_AB        (1<<4)	/* AB, Accept Broadcast */
#define   BBA_NCRB_HBD       (1<<5)	/* HBD, reserved */
#define   BBA_NCRB_RXINTC0   (1<<6)	/* RXINTC, Receive Interrupt Counter */
#define   BBA_NCRB_RXINTC1   (1<<7)	/*  " */
#define     BBA_NCRB_1_PACKET_PER_INT  (0<<6)	/* 0 0 */
#define     BBA_NCRB_2_PACKETS_PER_INT (1<<6)	/* 0 1 */
#define     BBA_NCRB_4_PACKETS_PER_INT (2<<6)	/* 1 0 */
#define     BBA_NCRB_8_PACKETS_PER_INT (3<<6)	/* 1 1 */

#define BBA_LTPS 0x04		/* Last Transmitted Packet Status, RO */
#define BBA_LRPS 0x05		/* Last Received Packet Status, RO */

#define BBA_IMR 0x08		/* Interrupt Mask Register, RW, 00h */
#define   BBA_IMR_FRAGIM     (1<<0)	/* FRAGIM, Fragment Counter Int Mask */
#define   BBA_IMR_RIM        (1<<1)	/* RIM, Receive Interrupt Mask */
#define   BBA_IMR_TIM        (1<<2)	/* TIM, Transmit Interrupt Mask */
#define   BBA_IMR_REIM       (1<<3)	/* REIM, Receive Error Interrupt Mask */
#define   BBA_IMR_TEIM       (1<<4)	/* TEIM, Transmit Error Interrupt Mask */
#define   BBA_IMR_FIFOEIM    (1<<5)	/* FIFOEIM, FIFO Error Interrupt Mask */
#define   BBA_IMR_BUSEIM     (1<<6)	/* BUSEIM, BUS Error Interrupt Mask */
#define   BBA_IMR_RBFIM      (1<<7)	/* RBFIM, RX Buffer Full Interrupt Mask */

#define BBA_IR 0x09		/* Interrupt Register, RW, 00h */
#define   BBA_IR_FRAGI       (1<<0)	/* FRAGI, Fragment Counter Interrupt */
#define   BBA_IR_RI          (1<<1)	/* RI, Receive Interrupt */
#define   BBA_IR_TI          (1<<2)	/* TI, Transmit Interrupt */
#define   BBA_IR_REI         (1<<3)	/* REI, Receive Error Interrupt */
#define   BBA_IR_TEI         (1<<4)	/* TEI, Transmit Error Interrupt */
#define   BBA_IR_FIFOEI      (1<<5)	/* FIFOEI, FIFO Error Interrupt */
#define   BBA_IR_BUSEI       (1<<6)	/* BUSEI, BUS Error Interrupt */
#define   BBA_IR_RBFI        (1<<7)	/* RBFI, RX Buffer Full Interrupt */

#define BBA_BP   0x0a/*+0x0b*/	/* Boundary Page Pointer Register */
#define BBA_TLBP 0x0c/*+0x0d*/	/* TX Low Boundary Page Pointer Register */
#define BBA_TWP  0x0e/*+0x0f*/	/* Transmit Buffer Write Page Pointer Register */
#define BBA_TRP  0x12/*+0x13*/	/* Transmit Buffer Read Page Pointer Register */
#define BBA_RWP  0x16/*+0x17*/	/* Receive Buffer Write Page Pointer Register */
#define BBA_RRP  0x18/*+0x19*/	/* Receive Buffer Read Page Pointer Register */
#define BBA_RHBP 0x1a/*+0x1b*/	/* Receive High Boundary Page Pointer Register */

#define BBA_RXINTT    0x14/*+0x15*/	/* Receive Interrupt Timer Register */

#define BBA_NAFR_PAR0 0x20	/* Physical Address Register Byte 0 */
#define BBA_NAFR_PAR1 0x21	/* Physical Address Register Byte 1 */
#define BBA_NAFR_PAR2 0x22	/* Physical Address Register Byte 2 */
#define BBA_NAFR_PAR3 0x23	/* Physical Address Register Byte 3 */
#define BBA_NAFR_PAR4 0x24	/* Physical Address Register Byte 4 */
#define BBA_NAFR_PAR5 0x25	/* Physical Address Register Byte 5 */

#define BBA_NWAYC 0x30		/* NWAY Configuration Register, RW, 84h */
#define   BBA_NWAYC_FD       (1<<0)	/* FD, Full Duplex Mode */
#define   BBA_NWAYC_PS100    (1<<1)	/* PS100/10, Port Select 100/10 */
#define   BBA_NWAYC_ANE      (1<<2)	/* ANE, Autonegotiation Enable */
#define   BBA_NWAYC_ANS_RA   (0x01<<3)	/* ANS, Restart Autonegotiation */
#define   BBA_NWAYC_LTE      (1<<7)	/* LTE, Link Test Enable */

#define BBA_GCA 0x32		/* GMAC Configuration A Register, RW, 00h */
#define   BBA_GCA_ARXERRB    (1<<3)	/* ARXERRB, Accept RX packet with error */

#define BBA_MISC 0x3d		/* MISC Control Register 1, RW, 3ch */
#define   BBA_MISC_BURSTDMA  (1<<0)
#define   BBA_MISC_DISLDMA   (1<<1)

#define BBA_TXFIFOCNT 0x3e/*0x3f*/	/* Transmit FIFO Counter Register */
#define BBA_WRTXFIFOD 0x48/*-0x4b*/	/* Write TX FIFO Data Port Register */

#define BBA_MISC2 0x50		/* MISC Control Register 2, RW, 00h */
#define   BBA_MISC2_HBRLEN0  (1<<0)	/* HBRLEN, Host Burst Read Length */
#define   BBA_MISC2_HBRLEN1  (1<<1)	/*  " */
#define   BBA_MISC2_AUTORCVR (1<<7)	/* Auto RX Full Recovery */

#define BBA_RX_STATUS_BF      (1<<0)
#define BBA_RX_STATUS_CRC     (1<<1)
#define BBA_RX_STATUS_FAE     (1<<2)
#define BBA_RX_STATUS_FO      (1<<3)
#define BBA_RX_STATUS_RW      (1<<4)
#define BBA_RX_STATUS_MF      (1<<5)
#define BBA_RX_STATUS_RF      (1<<6)
#define BBA_RX_STATUS_RERR    (1<<7)

#define BBA_TX_STATUS_CC0     (1<<0)
#define BBA_TX_STATUS_CC1     (1<<1)
#define BBA_TX_STATUS_CC2     (1<<2)
#define BBA_TX_STATUS_CC3     (1<<3)
#define  BBA_TX_STATUS_CCMASK (0x0f)
#define BBA_TX_STATUS_CRSLOST (1<<4)
#define BBA_TX_STATUS_UF      (1<<5)
#define BBA_TX_STATUS_OWC     (1<<6)
#define BBA_TX_STATUS_OWN     (1<<7)
#define BBA_TX_STATUS_TERR    (1<<7)

#define BBA_TX_MAX_PACKET_SIZE 1518	/* 14+1500+4 */
#define BBA_RX_MAX_PACKET_SIZE 1536	/* 6 pages * 256 bytes */


/*
 * EXpansion Interface glue for the Broadband Adapter.
 *
 */
#define BBA_EXI_ID 0x04020200

#define BBA_EXI_CHANNEL 0
#define BBA_EXI_DEVICE  2
#define BBA_EXI_FREQ    5

#define BBA_CMD_IR_MASKALL  0x00
#define BBA_CMD_IR_MASKNONE 0xf8


/* lwip stuff */

/* Define those to better describe your network interface. */
#define IFNAME0 'e'
#define IFNAME1 'n'

typedef unsigned long u32;
typedef unsigned char u8;
typedef unsigned short u16;

struct bba_private {
	struct eth_addr *ethaddr;
	/* Add whatever per-interface state that is needed here. */
	
	struct stats_s
	{
		int rx_errors, rx_over_errors, rx_crc_errors,
			rx_fifo_errors, rx_length_errors, rx_frame_errors;
		int rx_bytes, rx_packets, tx_errors, collisions, tx_carrier_errors, 
			tx_fifo_errors, tx_window_errors;
	} stats;
	
	u32 msg_enable;
	u8 revid;
	u8 __0x04_init[2];
	u8 __0x05_init;
};

static const struct eth_addr ethbroadcast = {{0xffU,0xffU,0xffU,0xffU,0xffU,0xffU}};

/* Forward declarations. */
static int bba_reset(struct netif *dev);

struct netif *gcif_netif;
/*-----------------------------------------------------------------------------------*/

#include "console.h"
#include "time.h"

static void bba_cmd_ins(int reg, void *val, int len);
static void bba_cmd_outs(int reg, void *val, int len);
static void bba_ins(int reg, void *val, int len);
static void bba_outs(int reg, void *val, int len);

#define bba_select()   exi_select(BBA_EXI_CHANNEL, BBA_EXI_DEVICE, BBA_EXI_FREQ)
#define bba_deselect() exi_deselect(BBA_EXI_CHANNEL)

static inline void bba_cmd_ins_nosel(int reg, void *val, int len)
{
	u16 req;
	req = reg << 8;
	exi_write(BBA_EXI_CHANNEL, &req, sizeof(req));
	exi_read(BBA_EXI_CHANNEL, val, len);
}

static void bba_cmd_ins(int reg, void *val, int len)
{
	bba_select();
	bba_cmd_ins_nosel(reg, val, len);
	bba_deselect();
}

static inline void bba_cmd_outs_nosel(int reg, void *val, int len)
{
	u16 req;
	req = (reg << 8) | 0x4000;
	exi_write(BBA_EXI_CHANNEL, &req, sizeof(req));
	exi_write(BBA_EXI_CHANNEL, val, len);
}

static void bba_cmd_outs(int reg, void *val, int len)
{
	bba_select();
	bba_cmd_outs_nosel(reg, val, len);
	bba_deselect();
}

static inline u8 bba_cmd_in8(int reg)
{
	u8 val;
	bba_cmd_ins(reg, &val, sizeof(val));
	return val;
}

static u8 bba_cmd_in8_slow(int reg)
{
	u8 val;
	bba_select();
	bba_cmd_ins_nosel(reg, &val, sizeof(val));
	udelay(200);
	bba_deselect();
	return val;
}

static inline void bba_cmd_out8(int reg, u8 val)
{
	bba_cmd_outs(reg, &val, sizeof(val));
}

static inline u8 bba_in8(int reg)
{
	u8 val;
	bba_ins(reg, &val, sizeof(val));
	return val;
}

static inline void bba_out8(int reg, u8 val)
{
	bba_outs(reg, &val, sizeof(val));
}

#define bba_in12(reg)     ((bba_in8(reg)&0xff)|((bba_in8((reg)+1)&0x0f)<<8))
#define bba_out12(reg,val) do { bba_out8((reg),(val)&0xff); \
	                        bba_out8((reg)+1,((val)&0x0f00)>>8); } while(0)

static inline void bba_ins_nosel(int reg, void *val, int len)
{
	u32 req;
	req = (reg << 8) | 0x80000000;
	exi_write(BBA_EXI_CHANNEL, &req, sizeof(req));
	exi_read(BBA_EXI_CHANNEL, val, len);
}

static void bba_ins(int reg, void *val, int len)
{
	bba_select();
	bba_ins_nosel(reg, val, len);
	bba_deselect();
}

static inline void bba_outs_nosel(int reg, void *val, int len)
{
	u32 req;
	req = (reg << 8) | 0xC0000000;
	exi_write(BBA_EXI_CHANNEL, &req, sizeof(req));
	exi_write(BBA_EXI_CHANNEL, val, len);
}

static void bba_outs(int reg, void *val, int len)
{
	bba_select();
	bba_outs_nosel(reg, val, len);
	bba_deselect();
}


#define bba_printf(format, arg...) \
        printf(format , ## arg)

/**
 *
 * DRIVER NOTES
 *
 * 1. Packet Memory organization
 *
 * rx: 15 pages of 256 bytes, 2 full sized packets only (6 pages each)
 * tx: through FIFO, not using packet memory
 *
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |1|2|3|4|5|6|7|8|9|A|B|C|D|E|F|
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * ^                           ^
 * |                           |
 * TLBP                        RHBP
 * BP
 *
 */

#define BBA_INIT_TLBP	0x00
#define BBA_INIT_BP	0x01
#define BBA_INIT_RHBP	0x0f
#define BBA_INIT_RWP	BBA_INIT_BP
#define BBA_INIT_RRP	BBA_INIT_BP

#define __BIG_ENDIAN_BITFIELD
#if defined(__BIG_ENDIAN_BITFIELD)
#define X(a,b)  b,a
#else
#define X(a,b)  a,b
#endif

struct bba_descr {
	u32 X(X(next_packet_ptr:12, packet_len:12), status:8);
} __attribute((packed));

static void bba_event_handler(int channel, int event, void *dev0);

static void gc_init(struct netif *netif)
{
	struct bba_private *priv;
	
	priv = (struct bba_private*)netif->state;
	
	printf("initializing BBA...\n");
	bba_reset(netif);
	exi_request_irq(2, EXI_EVENT_IRQ, bba_event_handler, netif);
	bba_event_handler(2, EXI_EVENT_IRQ, netif);
	
	printf("ok!\n");
	
	gcif_netif = netif;
}

/*-----------------------------------------------------------------------------------*/
/*
 * low_level_output():
 *
 * Should do the actual transmission of the packet. The packet is
 * contained in the pbuf that is passed to the function. This pbuf
 * might be chained.
 *
 */
/*-----------------------------------------------------------------------------------*/

static err_t
bba_output(struct netif *netif, struct pbuf *p)
{
//	struct bba_private *gcif = (struct bba_private*)netif->state;

	if (p->tot_len > BBA_TX_MAX_PACKET_SIZE)
	{
		bba_printf("dropping large packet\n");
		return 0;
	}

	/* wait until the TXFIFO is free */
	int to = 10000;
	while (bba_in8(BBA_NCRA) & (BBA_NCRA_ST0 | BBA_NCRA_ST1)) { udelay(1); if (!--to) break; } // wait until no tx in progress
	if (!to)
	{
		bba_printf("transmit timeout - dropping packet\n");
		return 0;
	}
	
	/* tell the card about the length of this packet */
	bba_out12(BBA_TXFIFOCNT, p->tot_len);

	/* store the packet in the TXFIFO, including padding if needed */
	struct pbuf *q;
	for(q = p; q != NULL; q = q->next) 
	{
		bba_select();
		bba_outs_nosel(BBA_WRTXFIFOD, q->payload, q->len);
		bba_deselect();
	}

	if (p->tot_len < ETH_ZLEN) {
		bba_select();
		u8 pad[ETH_ZLEN];
		int pad_len = ETH_ZLEN - p->tot_len;
		memset(pad, 0, pad_len);
		bba_outs_nosel(BBA_WRTXFIFOD, pad, pad_len);
		bba_deselect();
	}
	
	/* tell the card to send the packet right now */
	bba_out8(BBA_NCRA, (bba_in8(BBA_NCRA) | BBA_NCRA_ST1) & ~BBA_NCRA_ST0);
	
	return ERR_OK;
}

static int bba_rx_err(u8 status, struct netif *dev)
{
	struct bba_private *priv = (struct bba_private*)dev->state;
	int errors = 0;

	if (status & BBA_RX_STATUS_RERR) {
		errors++;
		priv->stats.rx_errors++;
		if (status & BBA_RX_STATUS_BF) {
			priv->stats.rx_over_errors++;
		}
		if (status & BBA_RX_STATUS_CRC) {
			priv->stats.rx_crc_errors++;
		}
		if (status & BBA_RX_STATUS_FO) {
			priv->stats.rx_fifo_errors++;
		}
		if (status & BBA_RX_STATUS_RW) {
			priv->stats.rx_length_errors++;
		}
		if (status & BBA_RX_STATUS_FAE) {
			priv->stats.rx_frame_errors++;
		}
	}
	if (status & BBA_RX_STATUS_RF) {
		priv->stats.rx_length_errors++;
	}

	if (errors) {
			bba_printf("rx errors, status %8.8x.\n",
					 status);

		/* stop receiver */
		bba_out8(BBA_NCRA, bba_in8(BBA_NCRA) & ~BBA_NCRA_SR);

		/* initialize page pointers */
		bba_out12(BBA_RWP, BBA_INIT_RWP);
		bba_out12(BBA_RRP, BBA_INIT_RRP);

		/* start receiver again */
		bba_out8(BBA_NCRA, bba_in8(BBA_NCRA) | BBA_NCRA_SR);
	}
	return errors;
}


/*-----------------------------------------------------------------------------------*/
/*
 * low_level_input():
 *
 * Should allocate a pbuf and transfer the bytes of the incoming
 * packet from the interface into the pbuf.
 *
 */
/*-----------------------------------------------------------------------------------*/
u8_t buffer[1500];

void bba_rx(struct netif *dev, int budget)
{
	struct bba_private *priv=(struct bba_private*)dev->state;
	struct bba_descr descr;
	struct pbuf *p, *q;
	int status, size;
	unsigned long pos, top;
	unsigned short rrp, rwp;
	int received = 0;

	/* get current receiver pointers */
	rwp = bba_in12(BBA_RWP);
	rrp = bba_in12(BBA_RRP);
	
	while ((received < budget) && (rrp != rwp))
	{
		unsigned char buf[4];
		bba_ins(rrp << 8, buf, sizeof(descr));
		*(u32*)&descr = buf[0] | (buf[1]<<8) | (buf[2]<<16) | (buf[3]<<24);
		
		size = descr.packet_len - 4;	/* ignore CRC */
		status = descr.status;

		/* abort processing in case of errors */
		if ((size > BBA_RX_MAX_PACKET_SIZE + 4)
				|| (status & (BBA_RX_STATUS_RERR | BBA_RX_STATUS_FAE))) {
			printf("Size %d, status %d\n", size, status);
			bba_rx_err(status, dev);
			continue;
		}

		pos = (rrp << 8) + 4;	/* skip descriptor */
		top = (BBA_INIT_RHBP + 1) << 8;

		if ((pos + size) < top) {
			/* full packet in one chunk */
			bba_ins(pos, buffer, size);
		} else {
			/* packet wrapped */
			int chunk_size = top - pos;

			bba_ins(pos, buffer, chunk_size);
			rrp = BBA_INIT_RRP;
			bba_ins(rrp << 8, buffer + chunk_size,
				size - chunk_size);
		}
		
		priv->stats.rx_bytes += size;
		priv->stats.rx_packets++;
		received++;
		
		int i;
		i=0;
		p = pbuf_alloc(PBUF_LINK, size, PBUF_POOL);
		if(p != NULL) {
			/* We iterate over the pbuf chain until we have read the entire
				 packet into the pbuf. */
			for(q = p; q != NULL; q = q->next) {
				/* Read enough bytes to fill this pbuf in the chain. The
					 avaliable data in the pbuf is given by the q->len
					 variable. */
				/* read data into(q->payload, q->len); */
				memcpy(q->payload, buffer+i, q->len);
				i += q->len;
			}
		}

		bba_out12(BBA_RRP, descr.next_packet_ptr);

		rwp = bba_in12(BBA_RWP);
		rrp = bba_in12(BBA_RRP);

		if (!p) /* out of memory or so */
			continue;

		struct eth_hdr *ethhdr = NULL;

		ethhdr = p->payload;
		
		q = NULL;
		switch (htons(ethhdr->type)) {
		/* IP packet? */
		case ETHTYPE_IP:
			/* update ARP table, obtain first queued packet */
			q = etharp_ip_input(dev, p);
			/* skip Ethernet header */
			pbuf_header(p, -14);
			LWIP_DEBUGF(NETIF_DEBUG, ("gc_input: passing packet up to IP\n"));
			/* pass to network layer */
			dev->input(p, dev);
			break;
		/* ARP packet? */
		case ETHTYPE_ARP:
			/* pass p to ARP module, get ARP reply or ARP queued packet */
			q = etharp_arp_input(dev, (struct eth_addr *)&dev->hwaddr, p);
			break;
		/* unsupported Ethernet packet type */
		default:
			/* free pbuf */
			pbuf_free(p);
			p = NULL;
			break;
		}
		/* send out the ARP reply or ARP queued packet */
		if (q != NULL) {
			/* q pbuf has been succesfully sent? */
			if (bba_output(dev, q) == ERR_OK)
			{
				pbuf_free(q);
				q = NULL;
			} else
			{
				/* TODO: re-queue packet in the ARP cache here (?) */
				pbuf_free(q);
				q = NULL;
			}
		}
	}
}

/**
 *
 */
static int bba_tx_err(u8 status, struct netif *dev)
{
	struct bba_private *priv=(struct bba_private*)dev->state;
	int errors = 0;

	if (status & BBA_TX_STATUS_TERR) {
		errors++;
		priv->stats.tx_errors++;
		if (status & BBA_TX_STATUS_CCMASK) {
			priv->stats.collisions +=
					(status & BBA_TX_STATUS_CCMASK);
		}
		if (status & BBA_TX_STATUS_CRSLOST) {
			priv->stats.tx_carrier_errors++;
		}
		if (status & BBA_TX_STATUS_UF) {
			priv->stats.tx_fifo_errors++;
		}
		if (status & BBA_TX_STATUS_OWC) {
			priv->stats.tx_window_errors++;
		}
	}

	if (errors) {
			bba_printf("tx errors, status %8.8x.\n",
				 status);
	}
	return errors;
}

static void inline bba_interrupt(struct netif *dev)
{
	struct bba_private *priv=(struct bba_private*)dev->state;
	
	unsigned long flags;
	u8 ir, imr, status;
	int loops = 0;

	ir = bba_in8(BBA_IR);
	imr = bba_in8(BBA_IMR);
	status = ir & imr;

	while (status) {
		bba_out8(BBA_IR, status);

		if (status & BBA_IR_RBFI) {
			bba_rx(dev, 0x0f);
		}
		if (status & BBA_IR_RI) {
			bba_rx(dev, 0x0f);
		}
		if (status & BBA_IR_REI) {
			printf("REI (status: %08x!\n", status);
			bba_rx_err(bba_in8(BBA_LRPS), dev);
		}
		if (status & BBA_IR_TI) {
			bba_tx_err(bba_in8(BBA_LTPS), dev);
			/* allow more packets to be sent */
		}
		if (status & BBA_IR_TEI) {
			bba_tx_err(bba_in8(BBA_LTPS), dev);
			/* allow more packets to be sent */
		}
		if (status & BBA_IR_FIFOEI) {
			/* allow more packets to be sent */
			printf("fifoei\n");
		}
		if (status & BBA_IR_BUSEI) {
			printf("busy\n");
		}
		if (status & BBA_IR_FRAGI) {
			printf("fragi\n");
		}

		ir = bba_in8(BBA_IR);
		imr = bba_in8(BBA_IMR);
		status = ir & imr;
		loops++;
	}
//	if (loops > 10)
//		bba_printf("a lot of interrupt work (%d loops)\n", loops);
}

err_t gcif_output(struct netif *dev, struct pbuf *p, struct ip_addr *ipaddr)
{
//	struct bba_private *gcif = dev->state;
	p = etharp_output(dev, ipaddr, p);
	/* network hardware address obtained? */
	if (p != NULL)
	{
	/* send out the packet */
		bba_output(dev, p);
		p = NULL;
	}
	// { p == NULL }
	else
	{
		printf(" DO NOT SEND PACKET.\n");
		/* we cannot tell if the packet was sent, the packet could have been queued */
		/* on an ARP entry that was already pending. */
	}
	return ERR_OK;
}

static int bba_reset(struct netif *dev)
{
	struct bba_private *priv=(struct bba_private*)dev->state;
	
	/* unknown, mx register 0x60 */
	bba_out8(0x60, 0);
	udelay(1000);

	/* unknown, command register 0x0f */
	bba_cmd_in8_slow(0x0f);
	udelay(1000);

	/* software reset (write 1 then write 0) */
	bba_out8(BBA_NCRA, BBA_NCRA_RESET);
	udelay(100);
	bba_out8(BBA_NCRA, 0);

	/* unknown, command register 0x01 */
	/* XXX obtain bits needed for challenge/response calculation later */
	priv->revid = bba_cmd_in8(0x01);

	/* unknown, command registers 0x04, 0x05 */
	bba_cmd_outs(0x04, priv->__0x04_init, 2);
	bba_cmd_out8(0x05, priv->__0x05_init);

	/* unknown, mx registers 0x5b, 0x5c, 0x5e */
	bba_out8(0x5b, bba_in8(0x5b) & ~(1 << 7));
	bba_out8(0x5e, 1);
	bba_out8(0x5c, bba_in8(0x5c) | 4);

	udelay(1000);

	/* accept broadcast, assert int for every two packets received */
	bba_out8(BBA_NCRB, BBA_NCRB_AB | BBA_NCRB_2_PACKETS_PER_INT);

	/* setup receive interrupt time out, in 40ns units */
	bba_out8(BBA_RXINTT, 0x00);
	bba_out8(BBA_RXINTT+1, 0x06); /* 0x0600 = 61us */

	/* auto RX full recovery */
	bba_out8(BBA_MISC2, BBA_MISC2_AUTORCVR);

	/* initialize packet memory layout */
	bba_out12(BBA_TLBP, BBA_INIT_TLBP);
	bba_out12(BBA_BP, BBA_INIT_BP);
	bba_out12(BBA_RHBP, BBA_INIT_RHBP);

	/* set receive page pointers */
	bba_out12(BBA_RWP, BBA_INIT_RWP);
	bba_out12(BBA_RRP, BBA_INIT_RRP);

	/* start receiver */
	bba_out8(BBA_NCRA, BBA_NCRA_SR);

	/* packet memory won't contain packets with RW, FO, CRC errors */
	bba_out8(BBA_GCA, BBA_GCA_ARXERRB);

	/* retrieve MAC address */
	bba_ins(BBA_NAFR_PAR0, dev->hwaddr, 6);

	/* clear all interrupts */
	bba_out8(BBA_IR, 0xFF);

	/* enable all interrupts */
	bba_out8(BBA_IMR, 0xFF & ~BBA_IMR_FIFOEIM);

	/* unknown, short command registers 0x02 */
	/* XXX enable interrupts on the EXI glue logic */
	bba_cmd_out8(0x02, BBA_CMD_IR_MASKNONE);

	/* DO NOT clear interrupts on the EXI glue logic !!! */
	/* we need that initial interrupts for the challenge/response */

	return 0;		/* OK */
}

/**
 *
 */
unsigned long bba_calc_response(unsigned long val, struct bba_private *priv)
{
	u8 revid_0, revid_eth_0, revid_eth_1;
	revid_0 = priv->revid;
	revid_eth_0 = priv->__0x04_init[0];
	revid_eth_1 = priv->__0x04_init[1];

	u8 i0, i1, i2, i3;
	i0 = (val & 0xff000000) >> 24;
	i1 = (val & 0x00ff0000) >> 16;
	i2 = (val & 0x0000ff00) >> 8;
	i3 = (val & 0x000000ff);

	u8 c0, c1, c2, c3;
	c0 = ((i0 + i1 * 0xc1 + 0x18 + revid_0) ^ (i3 * i2 + 0x90)
			) & 0xff;
	c1 = ((i1 + i2 + 0x90) ^ (c0 + i0 - 0xc1)
			) & 0xff;
	c2 = ((i2 + 0xc8) ^ (c0 + ((revid_eth_0 + revid_0 * 0x23) ^ 0x19))
			) & 0xff;
	c3 = ((i0 + 0xc1) ^ (i3 + ((revid_eth_1 + 0xc8) ^ 0x90))
			) & 0xff;

	return ((c0 << 24) | (c1 << 16) | (c2 << 8) | c3);
}

/**
 *
 */
static void bba_event_handler(int channel, int event, void *dev0)
{
	struct netif *dev = (struct netif*)dev0;
	struct bba_private *priv=(struct bba_private*)dev->state;
	
	
	unsigned long flags;
	register u8 status, mask;

	/* get interrupt status from EXI glue */
	status = bba_cmd_in8(0x03);

	/* XXX mask all EXI glue interrupts */
	bba_cmd_out8(0x02, BBA_CMD_IR_MASKALL);

	/* start with the usual case */
	mask = (1<<7);

	/* normal interrupt from the macronix chip */
	if (status & mask) {
		/* call our interrupt handler */
		bba_interrupt(dev);
		goto out;
	}
	
	/* "killing" interrupt, try to not get one of these! */
	mask >>= 1;
	if (status & mask) {
		bba_printf("bba: killing interrupt!\n");
		/* reset the adapter so that we can continue working */
		bba_reset(dev);
		goto out;
	}

	/* command error interrupt, haven't seen one yet */
	mask >>= 1;
	if (status & mask) {
		goto out;
	}

	/* challenge/response interrupt */
	mask >>= 1;
	if (status & mask) {
		unsigned long response;
		unsigned long challenge;

		/* kids, don't do it without an adult present */
		bba_cmd_out8(0x05, priv->__0x05_init);
		bba_cmd_ins(0x08, &challenge, sizeof(challenge));
		response = bba_calc_response(challenge, priv);
		bba_cmd_outs(0x09, &response, sizeof(response));

		goto out;
	}
	
	/* challenge/response status interrupt */
	mask >>= 1;
	if (status & mask) {
		/* better get a "1" here ... */
		u8 result = bba_cmd_in8(0x0b);
		if (result != 1) {
			bba_printf("challenge failed! (result=%d)\n", result);
		}
		goto out;
	}

	/* should not happen, treat as normal interrupt in any case */
	bba_printf("bba: unknown interrupt type = %d\n", status);
out:
	/* assert interrupt */
	bba_cmd_out8(0x03, mask);

	/* enable interrupts again */
	bba_cmd_out8(0x02, BBA_CMD_IR_MASKNONE);
}

err_t gcif_init(struct netif *dev)
{
	struct bba_private *priv;
	
		/* scan for bba */
	u8 val;
	bba_select();
	unsigned long cid;
	exi_write(BBA_EXI_CHANNEL, "\0", 2);
	exi_read(BBA_EXI_CHANNEL, &cid, 4);
	bba_deselect();
	
	if (cid != 0x04020200)
	{
		printf(" - no BBA detected.\n");
		return -1;
	}

	priv = mem_malloc(sizeof(struct bba_private));
	if (priv == NULL)
	{
		LWIP_DEBUGF(NETIF_DEBUG, ("csgc_init: out of memory for gcif priv\n"));
		return ERR_MEM;
	}
	// initialize lwip network interface
	dev->name[0] = IFNAME0;
	dev->name[1] = IFNAME1;
	/* downward functions */
	dev->output = gcif_output;
	dev->linkoutput = bba_output;

	// initialize gc specific interface state data pointer
	dev->state = priv;

	/* maximum transfer unit */
	dev->mtu = 1500;

	/* broadcast capability */
	dev->flags = NETIF_FLAG_BROADCAST;
	
	priv->ethaddr = (struct eth_addr *)&(dev->hwaddr[0]);
	priv->revid = 0xf0;
	priv->__0x04_init[0] = 0xd1;
	priv->__0x04_init[1] = 0x07;
	priv->__0x05_init = 0x4e;

	/* hardware address length */
	dev->hwaddr_len = 6;
	
	// intialize the gc chip
	gc_init(dev);
	printf("BBA found! hwaddr is %02x:%02x:%02x:%02x:%02x:%02x!\n",
		dev->hwaddr[0], dev->hwaddr[1], dev->hwaddr[2],
		dev->hwaddr[3], dev->hwaddr[4], dev->hwaddr[5]);
	return ERR_OK;
}
